--------------------------------------------------------------------------------
-- Company: 		Trenz Electronic
-- Engineer: 		Oleksandr Kiyenko
--------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
--------------------------------------------------------------------------------
entity sinc3 is
generic (
	C_IMPLEMENTATION 		: integer range 0 to 1	:= 0
);
port (
	clk						: in  STD_LOGIC := '0';
	clk_en					: in  STD_LOGIC	:= '1';
	mdat					: in  STD_LOGIC := '0';
	dec_rate				: in  STD_LOGIC_VECTOR(15 downto 0) := x"0100";
	-- Result
	tdata					: out STD_LOGIC_VECTOR(15 downto 0);
	tvalid					: out STD_LOGIC
);
end sinc3;
--------------------------------------------------------------------------------
architecture RTL of sinc3 is
--------------------------------------------------------------------------------
signal acc1				: UNSIGNED(36 downto 0)	:= (others => '0'); 
signal acc2				: UNSIGNED(36 downto 0)	:= (others => '0'); 
signal acc3				: UNSIGNED(36 downto 0)	:= (others => '0'); 
signal acc3_d2			: UNSIGNED(36 downto 0)	:= (others => '0'); 
signal diff1			: UNSIGNED(36 downto 0)	:= (others => '0'); 
signal diff2			: UNSIGNED(36 downto 0)	:= (others => '0'); 
signal diff3			: UNSIGNED(36 downto 0)	:= (others => '0'); 
signal diff3a			: UNSIGNED(36 downto 0)	:= (others => '0'); 
signal diff1_d			: UNSIGNED(36 downto 0)	:= (others => '0'); 
signal diff2_d			: UNSIGNED(36 downto 0)	:= (others => '0'); 
signal word_count		: UNSIGNED(15 downto 0)	:= (others => '0'); 
signal word_clk			: STD_LOGIC := '0';
signal mdat_i			: STD_LOGIC := '0';
--------------------------------------------------------------------------------
begin
--------------------------------------------------------------------------------
process(clk)
begin
	if(clk = '1' and clk'event)then
		mdat_i	<= mdat;
		if(clk_en = '1')then
			if(mdat_i = '1')then
				acc1 <= acc1 + 1;
			end if;
			acc2 <= acc2 + acc1;  
			acc3 <= acc3 + acc2;  
		end if;
	end if;
end process;

-- decimation stage (MCLKOUT/WORD_CLK)
process(clk)
begin
	if(clk = '1' and clk'event)then
		if(clk_en = '1')then
			if(word_count = (UNSIGNED(dec_rate)-1))then
				word_count	<= (others => '0');
			else
				word_count	<= word_count + 1;
			end if;
		end if;
	end if;
end process;

process(clk)
begin
	if(clk = '1' and clk'event)then
		if(clk_en = '1')then
			if(word_count = (resize(UNSIGNED(dec_rate(15 downto 1)),16)-1))then
				word_clk	<= '1';
			else
				word_clk	<= '0';
			end if;
		end if;
	end if;
end process;

-- Differentiator (including decimation stage) 
-- Perform the differentiation stage (FIR) at a lower speed. 
-- Z = one sample delay WORD_CLK = output word rate
--process(word_clk)
process(clk)
begin
	--if(word_clk = '1' and word_clk'event)then
	if(clk = '1' and clk'event)then
		if((clk_en = '1') and (word_clk = '1'))then
			diff1		<= acc3 - acc3_d2;  
			diff2		<= diff1 - diff1_d;  
			diff3		<= diff2 - diff2_d;  
			diff3a		<= diff2 - diff2_d;  
			acc3_d2 	<= acc3;  
			diff1_d 	<= diff1;  
			diff2_d 	<= diff2;   
		end if;
	end if;
end process;


-- Clock the Sinc output into an output register WORD_CLK = output word rate
--process(word_clk)
process(clk)
begin
	--if(word_clk = '1' and word_clk'event)then
	if(clk = '1' and clk'event)then
		if((clk_en = '1') and (word_clk = '1'))then
			tvalid		<= '1';
			case dec_rate is
				when x"0020" => 	-- 32
					if(diff3(15 downto 0) = TO_UNSIGNED(32768,16))then
						tdata <= STD_LOGIC_VECTOR(TO_UNSIGNED(65535,16));
					else
						tdata <= STD_LOGIC_VECTOR(diff3a(14 downto 0) & '0');
					end if;
				when x"0040" => 	-- 64
					if(diff3(18 downto 2) = TO_UNSIGNED(65536,17))then
						tdata <= STD_LOGIC_VECTOR(TO_UNSIGNED(65535,16));
					else
						tdata <= STD_LOGIC_VECTOR(diff3a(17 downto 2));
					end if;
				when x"0080" => 	-- 128
					if(diff3(21 downto 5) = TO_UNSIGNED(65536,17))then
						tdata <= STD_LOGIC_VECTOR(TO_UNSIGNED(65535,16));
					else
						tdata <= STD_LOGIC_VECTOR(diff3a(20 downto 5));
					end if;
				when x"0100" => 	-- 256
					if(diff3(24 downto 8) = TO_UNSIGNED(65536,17))then
						tdata <= STD_LOGIC_VECTOR(TO_UNSIGNED(65535,16));
					else
						tdata <= STD_LOGIC_VECTOR(diff3a(23 downto 8));
					end if;
				when x"0200" => 	-- 512
					if(diff3(27 downto 11) = TO_UNSIGNED(65536,17))then
						tdata <= STD_LOGIC_VECTOR(TO_UNSIGNED(65535,16));
					else
						tdata <= STD_LOGIC_VECTOR(diff3a(26 downto 11));
					end if;
				when x"0400" => 	-- 1024
					if(diff3(30 downto 14) = TO_UNSIGNED(65536,17))then
						tdata <= STD_LOGIC_VECTOR(TO_UNSIGNED(65535,16));
					else
						tdata <= STD_LOGIC_VECTOR(diff3a(29 downto 14));
					end if;
				when x"0800" => 	-- 2048
					if(diff3(33 downto 17) = TO_UNSIGNED(65536,17))then
						tdata <= STD_LOGIC_VECTOR(TO_UNSIGNED(65535,16));
					else
						tdata <= STD_LOGIC_VECTOR(diff3a(32 downto 17));
					end if;
				when x"1000" => 	-- 4096
					if(diff3(36 downto 20) = TO_UNSIGNED(65536,17))then
						tdata <= STD_LOGIC_VECTOR(TO_UNSIGNED(65535,16));
					else
						tdata <= STD_LOGIC_VECTOR(diff3a(35 downto 20));
					end if;
				when others => 
					if(diff3(24 downto 8) = TO_UNSIGNED(65536,17))then
						tdata <= STD_LOGIC_VECTOR(TO_UNSIGNED(65535,16));
					else	
						tdata <= STD_LOGIC_VECTOR(diff3a(23 downto 8));
					end if;
			end case;
		else
			tvalid		<= '0';
		end if;
	end if;
end process;
--------------------------------------------------------------------------------
end RTL;
